--==============================================================================
-- CERN (BE-CO-HT)
-- Burst mode control module
-- Copyright CERN 2017
--==============================================================================
--
-- author: Denia Bouhired (denia.bouhired@cern.ch)
--
-- Date of creation: 19-09-2016 
--
-- version: 1.0
--
-- Description:
-- This module serves as a burst mode controller. When pulses of  
-- pre-defined length (250 ns) arrive, depending on the frequency, the 
-- module will allow the pulse to go through for a pre-defined amount of
-- time, before going into pulse rejection mode. The rejection lasts for
-- the time it takes for the "temperature" to go below the set upper limit
 -- g_max_temp. For each frequency, the time of failure selected
-- corresponds to the time it takes to reach g_max_temp for pulses at the 
-- same frequency. This is mapped onto the temperature rise caused by a 
-- single pulse (the maximum rise at 2MHz max frequency is 
-- g_1_pulse_temp_rise) of the corresponding period. The array of values
-- representing the thermal properties at the pulse level is given as the 
-- array of integers temp_decre_step. This array of values is 
-- generated in pre-processing via python script (Link??). These values 
-- correspond to the thermal model of the board components defined at 
-- the pulse level. 

-- Any modification to the board specification which would change the 
-- high frequency operation behaviour, would require changing the 3
-- parameters g_1_pulse_temp_rise, g_max_temp and t_temp_decre_step. These 
-- are generated using the Python file (link?) 




-- dependencies:
--    

--==============================================================================
-- GNU LESSER GENERAL PUBLIC LICENSE
--==============================================================================
-- This source file is free software; you can redistribute it and/or modify it
-- under the terms of the GNU Lesser General Public License as published by the
-- Free Software Foundation; either version 2.1 of the License, or (at your
-- option) any later version. This source is distributed in the hope that it
-- will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
-- of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
-- See the GNU Lesser General Public License for more details. You should have
-- received a copy of the GNU Lesser General Public License along with this
-- source; if not, download it from http://www.gnu.org/licenses/lgpl-2.1.html
--==============================================================================


library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.gencores_pkg.all;
use work.wishbone_pkg.all;
use work.conv_common_gw_pkg.all;

--============================================================================
--ENTITY DECLARATION
--============================================================================

entity conv_dyn_burst_ctrl is
	generic
	(
		-- Fixed pulse width set to 5 clock cycles = 5* 50ns = 250 ns
		g_pwidth 				: natural range 2 to 40 	:= 5;
		g_temp_decre_step 		: t_temp_decre_step 		:=
							(0, 769, 31, 104, 14, 82, 0 ,0, 0, 0, 0, 0, 0, 0, 0);
		
		--Scaled temperature rise resulting from single pulse.
		g_1_pulse_temp_rise		:in unsigned (19 downto 0)	:= x"01388"; --5000
		
		-- Scaled maximum temperature ceiling before pulse inhibition 
		-- is necessary to lower temperature
		g_max_temp				:in unsigned (39 downto 0) 	:= x"02540BE400" --10^10
	);
	port
	(
    -- Clock and active-low reset inputs
    clk_i        		: in  std_logic;
    rst_n_i      		: in  std_logic;

	-- Enable input, high frequency repetition is enabled when '1'
    en_i         	 	: in  std_logic;
	
	-- Asynchronous input pulse with rising and falling edges
	pulse_burst_i 		: in std_logic;
	pulse_r_edge_p_i	: in std_logic;
	pulse_f_edge_p_i	: in std_logic;
	
	-- Dynamic,temperature controlled ouput pulse train.
	pulse_burst_o		: out std_logic;
	
	-- Burst error output, pulses high for one clock cycle when a pulse arrives
	-- within a burst rejection phase
	burst_err_p_o : out std_logic	
	);
	end entity conv_dyn_burst_ctrl;
	
-----------------------------------------------------------------------------------------	
-- ARCHITECTURE
-----------------------------------------------------------------------------------------
	architecture behav of conv_dyn_burst_ctrl is
	
	--type t_temp_decre_step is array (0 to 5) of integer; 
	
	type t_state is (
		IDLE,			
		PULSE_REPEAT,	
		PULSE_REJECT	
	);

	--The following t_temp_decre_step values correspond to "1s, 6.5s, 10s, 26s, 36.66s and continuous" 
	--for pulsing for frequencies 2MHz, 1.33MHz, 1MHz, 800kHz, 667 kHz and 571kHz respectively.
	--signal temp_decre_step  : t_temp_decre_step := (0, 769, 31, 104, 14, 82);
	
	signal burst_ctrl_rst			: std_logic;
	signal pulse_train_in 			: std_logic;
	signal temp_rise				: unsigned (39 downto 0) ;
	signal single_cycle_cnt			: integer;
	signal n_cycle_cnt				: integer range 0 to g_temp_decre_step'LENGTH;
	signal pulse_train_in_d0 		: std_logic;
	signal pulse_train_in_r_edge_p  : std_logic;
	signal pulse_train_in_f_edge_p  : std_logic;
	signal thermal_array_lgth :natural := 6;
	
	signal state                     : t_state;
	signal nxt_state			     : t_state;
	constant thermal_res : natural := g_pwidth; -- thermal resolution in clock cycles
	
	begin
	
	thermal_array_lgth <= 6 when g_pwidth = 5 else 15;
	
    pulse_train_in <= '0' when burst_ctrl_rst = '1' else pulse_burst_i and en_i;
	pulse_burst_o <= pulse_train_in;

--Process to count in n clk cycles steps
	p_n_cycle_cnt : process(clk_i)
	begin
	if rising_edge(clk_i) then
		if rst_n_i = '0' then
				single_cycle_cnt <= 1;
				n_cycle_cnt   <= 1;
		else
			--reset counters in the event of a new pulse only when pulse rejection is not activ
			if (pulse_r_edge_p_i = '1' or pulse_f_edge_p_i = '1') and burst_ctrl_rst = '0' then	
				single_cycle_cnt <= 1;
				n_cycle_cnt   <= 1;
			--When no pulse is being repeated, whether because pulse is off or because it is being rejected
			else
				single_cycle_cnt <= single_cycle_cnt + 1; 	--count clk cycles 
				if single_cycle_cnt = thermal_res then
					if n_cycle_cnt < thermal_array_lgth then
						n_cycle_cnt <= n_cycle_cnt + 1; 	--increment every n=thermal_res clk cycles
					end if;
					single_cycle_cnt <= 1;
				end if;
			end if;
		end if;
	end if;
  end process p_n_cycle_cnt;

-----------------------------------------------------------------------------------------	
-- Finite State Machine FSM
-- Finite State Machine to control pulse repetition as a function of rising board temperature.
-- The FSM uses a temp_rise counter 
-----------------------------------------------------------------------------------------
  
-- Process to trigger state transitions
----------------------------------------
  p_fsm_transitions: process(clk_i)
  begin
    if rising_edge(clk_i) then
      if rst_n_i = '0' then
        state 						<= IDLE;
      elsif (en_i = '1') then
		state	 					<= nxt_state;
      end if;
    end if;
  end process;
  
  
-- Process to define FSM states
--------------------------------
  p_thermal_fsm_states : process (state, pulse_r_edge_p_i, pulse_f_edge_p_i, temp_rise, n_cycle_cnt )
  begin
	case state is 
	
	-----------------------------------------------------------------------------------------
	-- The FSM is IDLE, when the board is completely "cool", temp_rise =0, 
	-- and no new pulses arrive after 1.75us from the last one
	-----------------------------------------------------------------------------------------
		when IDLE =>
		
			if pulse_r_edge_p_i = '1' then
				if temp_rise >= 0 and temp_rise < g_max_temp then
					nxt_state 	<= PULSE_REPEAT;
				else
					nxt_state 	<= PULSE_REJECT; 
				end if;
				
			end if;
					
	-----------------------------------------------------------------------------------------
	-- PULSE_REPEAT pulses are repeated as long as the temperature is below 
	-- maximum g_max_temp.
	-- While the temperature counter temp_rise is above 0, the time between 
	-- 2 pulses is used to decrement it.
	-----------------------------------------------------------------------------------------
		when PULSE_REPEAT =>
		
			if temp_rise <= g_max_temp then
				if pulse_r_edge_p_i = '1' then
					nxt_state <= PULSE_REPEAT;
				elsif temp_rise = 0 and n_cycle_cnt = thermal_array_lgth then
					nxt_state <= IDLE;
				end if;
			else
				nxt_state <= PULSE_REJECT;
			end if;
			
	-----------------------------------------------------------------------------------------
	-- PULSE_REJECT applies when a new pulse causes temperature to exceed 
	-- maximum value
	-- i.e. temp_rise >= g_max_temp.
	-----------------------------------------------------------------------------------------				
		when PULSE_REJECT =>
			
			if temp_rise <= g_max_temp then
				if pulse_f_edge_p_i ='1' then
					nxt_state <= PULSE_REPEAT;
				end if;
			else
				nxt_state <= PULSE_REJECT;
			end if;

	end case;
			
	end process p_thermal_fsm_states;


-- Process to define FSM outputs
--------------------------------
	p_thermal_fsm_outputs : process (state,pulse_r_edge_p_i, pulse_f_edge_p_i, single_cycle_cnt, n_cycle_cnt)
  begin	    

	-----------------------------------------------------------------------------------------
	-- The FSM is IDLE, when the board is completely "cool", temp_rise =0, and no
	-- new pulses arrive after 1.75us from the last one
	-----------------------------------------------------------------------------------------	  
			case state is 
			
			when IDLE =>
				
				temp_rise 			<= (others => '0');
				burst_ctrl_rst 	 	<= '0';
				burst_err_p_o 		<= '0';

	-----------------------------------------------------------------------------------------
	-- PULSE_REPEAT pulses are repeated as long as the temperature is below maximum g_max_temp.
	-- While the temperature counter temp_rise is above 0, the time between 2 pulses is used to decrement it.
	-----------------------------------------------------------------------------------------
			
			
			when PULSE_REPEAT =>
				
					burst_ctrl_rst <= '0';
					burst_err_p_o <= '0';
					
					if pulse_f_edge_p_i = '1' then
						temp_rise <=  temp_rise + g_1_pulse_temp_rise;
					else
						if temp_rise >= g_temp_decre_step(n_cycle_cnt-1) then
							temp_rise <= temp_rise - to_unsigned(g_temp_decre_step(n_cycle_cnt-1), 40);
						else
							temp_rise <= (others => '0');
						end if;
					end if;

					
	-----------------------------------------------------------------------------------------
	-- PULSE_REJECT applies when a new pulse causes temperature to exceed maximum value
	--i.e. temp_rise >= g_max_temp.
	-----------------------------------------------------------------------------------------
			
			when PULSE_REJECT =>
					burst_ctrl_rst <= '1';
					burst_err_p_o <= pulse_r_edge_p_i;
					temp_rise <= temp_rise - g_temp_decre_step(n_cycle_cnt-1);
			end case;

			
	end process p_thermal_fsm_outputs;

	end architecture behav;